/*
  'Ridge' kwin client
  file: Ridge.cpp
  date: 08.12.2003

  (c) 2003 Jens Luetkens <j.luetkens@limitland.de>
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING.  If not, write to
  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
*/

#include <unistd.h> // for usleep
#include <math.h>

#include <qlayout.h>
#include <qpainter.h>
#include <qimage.h>

#include <netwm.h>

#undef Bool

#include <kwin/workspace.h>
#include <kwin/options.h>

#include <kconfig.h>
#include <kstandarddirs.h>
#include <kapplication.h>
#include <klocale.h>
#include <kpixmap.h>
#include <kpixmapeffect.h>

#include "Ridge.h"
#include "RidgeButton.h"



using namespace KWinInternal;

extern "C"
{
	Client * allocate(Workspace * ws, WId w, int tool)
	{
		return new Ridge(ws, w, tool != 0);
	}

	void init()
	{
		// Empty.
	}

	void reset()
	{
		Workspace::self()->slotResetAllClientsDelayed();
		// Empty.
	}

	void deinit()
	{
		// Empty.
	}
}

Ridge::Ridge(Workspace * ws, WId w, bool tool, QWidget * parent, const char * name)
		: Client        (ws, w, parent, name, WResizeNoErase),
		tool_			(tool),
		hasMenu			(false),
		mainLayout_		(0),
		topSpacer_	(0)
{
	setBackgroundMode(NoBackground);

	resetLayout();

	leftButtonList_   .setAutoDelete(true);
	rightButtonList_  .setAutoDelete(true);

	connect(options, SIGNAL(resetClients()), this, SLOT(slotReset()));

}

Ridge::~Ridge()
{
	// Empty.
}

//
// paintEvent(QPaintEvent *)
//
void Ridge::paintEvent(QPaintEvent *)
{
	QPainter p(this);
	QPainter	painter;
	QFontMetrics fm(options->font(isActive(), tool_));
	
	titleBarColor = options->color(Options::TitleBar, isActive());
	titleBlendColor = options->color(Options::TitleBlend, isActive());

	titleRect = topSpacer_->geometry();
	titleRect.setHeight(boxHeight);
	titleWidth = QMIN(fm.width(caption()), titleRect.width());
	titleWidth += textMargin + textMargin + textMargin;
	if ( hasMenu ) {
		titleWidth += textMargin + menuIcon.width();
	}
	titleRect.setWidth(titleWidth);

	// shapes
	if ( m_hasOutline )
	    tempPixmap.resize(1, boxHeight-1);
	else
	    tempPixmap.resize(1, boxHeight);
    KPixmapEffect::gradient(tempPixmap,
                   	        titleBarColor,
                   	        titleBlendColor,
                   	        KPixmapEffect::VerticalGradient);
    backgroundTile.resize(1, tempPixmap.height());
    painter.begin(&backgroundTile);
    painter.drawPixmap(0, 0, tempPixmap);
    painter.end();


	if ( isShade() ) {
	
		setMask(titleRect);
	
	}	
	else {

		QRect windowRect(rect());

		// window background
		if ( m_hasOffset )
			windowRect.setTop(m_offsetSize);
		QRegion mask(windowRect);
		if ( m_hasOutline ) {
			frameRidgeRect(windowRect, &p);
			insetRect(&windowRect, 1);
		}
		
		if ( m_decorationTheme == "Gradient" )
			p.fillRect(windowRect, titleBlendColor);
		else 
			decorateRidgeRect(windowRect, &p, titleBlendColor, true);


		// top Layout Background
		if ( m_hasOutline )
			windowRect.setHeight(boxHeight-1);
		else
			windowRect.setHeight(boxHeight);
		decorateRidgeRect(windowRect, &p, titleBlendColor, false);

		// Mask without buttons
		for (QPtrListIterator<RidgeButton> it(leftButtonList_); it.current(); ++it)
		{
			mask += it.current()->frameGeometry();
		}
		for (QPtrListIterator<RidgeButton> it(rightButtonList_); it.current(); ++it)
		{
			mask += it.current()->frameGeometry();
		}
		mask += titleRect;
		setMask(mask);

	}

	// title Rect
	if ( m_hasOutline ) {
		frameRidgeRect(titleRect, &p);
		insetRect(&titleRect, 1);
	}
	decorateRidgeRect(titleRect, &p, titleBarColor, true);
	insetRect(&titleRect, textMargin);
	
	
	// draw icon
	if ( hasMenu ) {
	    p.drawPixmap(titleRect.left(), titleRect.top(), menuIcon);
		titleRect.moveBy(menuIcon.width()+textMargin,0);
		titleRect.setWidth(titleRect.width()-menuIcon.width()-textMargin);
	}
	titleRect.setTop(titleRect.top()-(textMargin/2));
	titleRect.setHeight(titleRect.height()+(textMargin/2));

	//draw text
	p.setFont(options->font(isActive(), tool_));
	p.setPen(options->color(Options::Font, isActive()));
	p.drawText(titleRect, AlignLeft, caption());

}

//
// decorateRidgeRect( QRect rect, QPainter * p, QColor backgroundColor, bool fill )
//
void Ridge::decorateRidgeRect( QRect rect, QPainter * p, QColor backgroundColor, bool fill )
{
	if ( m_hasDecoration ) {
		if ( m_decorationTheme == "Gradient" ) {
			p->drawTiledPixmap(rect, backgroundTile);
		}
		else {
			QColor lightColor = backgroundColor.light(120);
			QColor darkColor = backgroundColor.dark(120);

			if ( fill )
				p->fillRect(rect, backgroundColor);
			
			//draw long bevelling lines
			p->setPen(darkColor);
			p->drawLine(rect.left()+1, rect.bottom(), rect.right(), rect.bottom());
			p->drawLine(rect.right(), rect.bottom(),rect.right(), rect.top()+1);
			p->drawLine(rect.left()+1, rect.bottom()-2, rect.left()+1, rect.top()+1);
			p->drawLine(rect.left()+1, rect.top()+1, rect.right()-2, rect.top()+1);
			p->setPen(lightColor);
			p->drawLine(rect.left(), rect.bottom()-1, rect.left(), rect.top());
			p->drawLine(rect.left(), rect.top(), rect.right()-1, rect.top());
			p->drawLine(rect.left()+2, rect.bottom()-1, rect.right()-1, rect.bottom()-1);
			p->drawLine(rect.right()-1, rect.bottom()-1,rect.right()-1, rect.top()+2);
		}
	}
	else {
		p->fillRect(rect, backgroundColor);
	}
}

//
// frameRidgeRect( QRect rect, QPainter * p 
//
void Ridge::frameRidgeRect( QRect rect, QPainter * p )
{

	if ( isActive() )
		p->setPen(m_outlineAColor);
	else
		p->setPen(m_outlineIColor);

	p->drawRect(rect);

}

//
// insetRect( QRect * rect, int inset )
//
void Ridge::insetRect( QRect * rect, int inset )
{

	rect->setWidth(rect->width()-inset-inset);
	rect->setHeight(rect->height()-inset-inset);
	rect->moveBy(inset,inset);

}

//
// mouseDoubleClickEvent(QMouseEvent * e)
//
void Ridge::mouseDoubleClickEvent(QMouseEvent * e)
{
	if (topSpacer_->geometry().contains(e->pos()))
	{
		workspace()->performWindowOperation(this, options->operationTitlebarDblClick());
	}
}

//
// stickyChange(bool b)
//
void Ridge::stickyChange(bool b)
{
	emit(stkyChange(b));
}

//
// maximizeChange(bool b)
//
void Ridge::maximizeChange(bool b)
{
	emit(maxChange(b));
}

//
// activeChange(bool)
//
void Ridge::activeChange(bool)
{
	repaint();
	repaintButtons();
}

//
// slotReset()
//
void Ridge::slotReset()
{
	resetLayout();
	repaint();
}

//
// createButton(const QString & s)
//
RidgeButton * Ridge::createButton(const QString & s)
{
	RidgeButton * b = 0;

	if (("Help" == s) && providesContextHelp()) {
		b = new RidgeButton(this, "help", i18n("Help"));
		connect(b, SIGNAL(clicked()), this, SLOT(contextHelp()));
	}
/*
	else if ("Menu" == s) {
		b = new RidgeButton(this, "menu", i18n("Menu"));
		connect(b, SIGNAL(pressed()), this, SLOT(menuButtonPressed()));
		connect(b, SIGNAL(released()), this, SLOT(menuButtonReleased()));
	}
*/
	else if ("Sticky" == s) {
		b = new RidgeButton(this, "sticky", i18n("Sticky"));
		connect(b, SIGNAL(clicked()), this, SLOT(toggleSticky()));
		connect(this, SIGNAL(stkyChange(bool)), b, SLOT(slotStickyChange(bool)));
	}
	else if ("Iconify" == s && isMinimizable()) {
		b = new RidgeButton(this, "iconify", i18n("Minimize"));
		connect(b, SIGNAL(clicked()), this, SLOT(iconify()));
	}
	else if ("Maximize" == s && isMaximizable()) {
		b = new RidgeButton(this, "maximize", i18n("Maximize"));
		connect(b, SIGNAL(clicked()), this, SLOT(maximize()));
		connect(this, SIGNAL(maxChange(bool)), b, SLOT(slotMaximizeChange(bool)));
	}
	else if ("Close" == s) {
		b = new RidgeButton(this, "close", i18n("Close"));
		connect(b, SIGNAL(clicked()), this, SLOT(closeWindow()));
	}
	else if ("Lower" == s) {
	}

	return b;
}

//
// createButtons()
//
void Ridge::createButtons()
{
	leftButtonList_   .clear();
	rightButtonList_  .clear();

	QString buttons = options->titleButtonsLeft() + "|" + options->titleButtonsRight();
	QPtrList<RidgeButton> *buttonList = &leftButtonList_;
	for (unsigned int i = 0; i < buttons.length(); ++i)
	{
		RidgeButton * tb = 0;
		switch (buttons[i].latin1())
		{
		case 'H': // Help
			tb = createButton("Help");
			break;
		case 'M': // Menu
        	menuIcon = icon();
           	menuIcon.convertFromImage( menuIcon.convertToImage().smoothScale(fontHeight-1, fontHeight-1));
			hasMenu = true;
			break;
		case 'S': // Sticky
			tb = createButton("Sticky");
			break;
		case 'I': // Minimize
			tb = createButton("Iconify");
			break;
		case 'A': // Maximize
			tb = createButton("Maximize");
			break;
		case 'X': // Close
			tb = createButton("Close");
			break;
		case '|':
			buttonList = &rightButtonList_;
			break;
		}
		if (0 != tb)
			buttonList->append(tb);
	}

}

//
// someGeometry()
//
void Ridge::someGeometry() 
{
	
	QColor	defaultcolor(0, 0, 0);
	
    // read configuration
	KConfig c(locate("config", "kwinridgerc"));
	c.setGroup("General");
    m_hasBorder			= c.readBoolEntry("hasBorder", true);
    m_borderSize		= c.readNumEntry("borderSize", 3);
    m_hasOffset			= c.readBoolEntry("hasOffset", true);
	m_offsetSize		= c.readNumEntry("offsetSize", 5);
    m_hasDecoration		= c.readBoolEntry("hasDecoration", true);
	m_decorationTheme	= c.readEntry("decorationTheme", "Ridge");
    m_hasOutline	    = c.readBoolEntry("hasOutline", true);
	m_outlineAColor		= c.readColorEntry("outlineAColor", &defaultcolor);
	m_outlineIColor		= c.readColorEntry("outlineIColor", &defaultcolor);

	QFontMetrics fm(options->font(isActive(), tool_));
	
	textMargin = QMAX(3,fm.height() / 5);
	fontHeight = QMAX(8, fm.height());
	boxHeight = fontHeight + textMargin + textMargin;
	// must be a multiple of 2
	if (! boxHeight%2==0 )
		boxHeight += 1;
	if ( m_hasOffset ) {
		topHeight = boxHeight + m_offsetSize;
	}
	else {
		topHeight = boxHeight;
	}

}

//
// resetLayout()
//
void Ridge::resetLayout()
{

	someGeometry();

	// mainLayout:
	//  ____________________________________
	// |  |                              |  |
	// |Xo|     topSpacer                |v^| <--- topLayout
	// |__|______________________________|__|
	// | |                                | |
	// | |                                | |
	// | |     windowWrapper              | |
	// | |                                | | <--- midLayout
	// | |                                | |
	// | |                                | |
	// | |________________________________| |
	// |____________________________________|
	

	delete mainLayout_;
	mainLayout_		= new QVBoxLayout(this, 0, 0);

	topSpacer_		= new QSpacerItem(0,topHeight,QSizePolicy::Expanding,QSizePolicy::Fixed);

	QHBoxLayout * topLayout;
	if ( m_hasOffset )
		topLayout   = new QHBoxLayout(mainLayout_, m_offsetSize, 0);
	else 
		topLayout   = new QHBoxLayout(mainLayout_, 0, 0);
	
	createButtons();

	if ( m_hasOffset )
		topLayout->addSpacing(m_offsetSize);

	// Add left-side buttons.
	for (QPtrListIterator<RidgeButton> it(leftButtonList_); it.current(); ++it)
	{
		topLayout->addWidget(it.current(), 0, Qt::AlignTop);
		topLayout->setStretchFactor(it.current(), 0);
		it.current()->setFixedSize(boxHeight, boxHeight);
	}

	// the title spacer and space to the right
	topLayout->addItem(topSpacer_);
	topLayout->addSpacing(textMargin+textMargin+textMargin+textMargin);
	if ( m_hasOffset )
		topLayout->addSpacing(m_offsetSize);
	if ( hasMenu ) 
		topLayout->addSpacing(menuIcon.width());
	
	// Add right-side buttons.
	for (QPtrListIterator<RidgeButton> it(rightButtonList_); it.current(); ++it)
	{
		topLayout->addWidget(it.current(), 0, Qt::AlignTop);
		topLayout->setStretchFactor(it.current(), 0);
		it.current()->setFixedSize(boxHeight, boxHeight);
	}

	if ( m_hasOffset )
		topLayout->addSpacing(m_offsetSize);

	// -------------------------------------------------------------------

	QHBoxLayout * midLayout   = new QHBoxLayout(mainLayout_, 0, 0);

	if ( m_hasBorder )
		midLayout->addSpacing(m_borderSize);
	midLayout->addWidget(windowWrapper());
	if ( m_hasBorder )
		midLayout->addSpacing(m_borderSize);

	// -------------------------------------------------------------------

	if ( m_hasBorder )
		mainLayout_->addSpacing(m_borderSize);

	// Make sure that topLayout doesn't stretch - midLayout should take
	// all spare space.
	mainLayout_->setStretchFactor(topLayout, 0);
	mainLayout_->setStretchFactor(midLayout, 1);

}

//
// repaintButtons(void)
//
void Ridge::repaintButtons(void)
{
	//force all buttons to repaint
	QPtrListIterator<RidgeButton> leftIterator(leftButtonList_);
	RidgeButton *button = NULL;
	while ( (button = leftIterator.current()) != NULL)
	{
		button->repaint();
		++leftIterator;
	}

	QPtrListIterator<RidgeButton> rightIterator(rightButtonList_);
	while ( (button = rightIterator.current()) != NULL)
	{
		button->repaint();
		++rightIterator;
	}
}

